#ifndef TERRAIN_CPP
#define TERRAIN_CPP

#include "common.h"
#include "geometry.cpp"
#include "corner.cpp"
#include "texture.h"

#define LEFT  0
#define RIGHT  200000.0
#define FAR -200000.0
#define NEAR 0.0

#define MAX_HORZ_POINTS 300
#define MAX_VERT_POINTS 300

class Terrain	//class that generates and draws a terrain
{
public:
	TEXTURE texture[4];
	bool bDrawNormals;	//specifies whether to draw normals 
	bool bDoFlatShading;	//specifies whether to do flat shading
	bool bPerFaceNormals;	//specifies whether to use per-face or per vertex normals
	Corner allPoints[MAX_HORZ_POINTS][MAX_VERT_POINTS];	//array for points
	std::vector<Face> allFaces;	//vector that stores all faces
	int nHorizontalPoints;	//X granularity
	int nVerticalPoints;	//Z granularity
	double nBumpFraction;	//multiplier for height of each vertex when generating
	double faceLength;
	double faceWidth;
	double maxHeight;
	double minHeight;
	int terrainDL;
	bool dlSaved;
	double smoothingPower;
	int numSpheres;
	double minBubbleSize;
	double maxBubbleSize;
	Terrain()
	{
		bDrawNormals = false;
		bDoFlatShading = false;
		nHorizontalPoints = 100;
		nVerticalPoints = 100;	
		nBumpFraction = 1.4;	
		bPerFaceNormals = false;
		maxHeight = 25000;
		minHeight = 0;
		smoothingPower = 1.1;
		numSpheres = 800;
		dlSaved = false;
		maxBubbleSize = (NEAR - FAR) / 8;
		minBubbleSize = 500;
		
		srand(time(NULL));

		regenerate();
	}
	~Terrain()
	{
		
	}
	double heightFunction(double x,double z)
	{
		if(x < LEFT || x > RIGHT || z < FAR || z > NEAR)
		{
			return 0;
		}
		int col = (int)floor((x - LEFT)/ faceLength);
		int row = (int)floor((z - FAR)/faceWidth);
		double fractX = (x - LEFT)/ faceLength - col;
		double fractZ = (z - FAR)/faceWidth - row;
		Vertex * v1;
		Vertex * v2;
		Vertex * v3;
		if(fractX + fractZ >= 1)
		{
			v1 = &allPoints[col + 1][row];
			v2 = &allPoints[col + 1][row + 1];
			v3 = &allPoints[col][row + 1];
		}
		else
		{
			v1 = &allPoints[col + 1][row];
			v2 = &allPoints[col + 1][row + 1];
			v3 = &allPoints[col][row + 1];
		}
		Arrow a1 = v1->sub(v2);
		Arrow a2 = v3->sub(v2);
		Arrow normal = a1.cross(a2);
		double r= normal.dot(v1);
		return (r - x*normal.getX() - z*normal.getZ())/normal.getY();		
	}
	Face * getFace(double x,double z)
	{
			if(x < LEFT || x > RIGHT || z < FAR || z > NEAR)
			{
				return 0;		
			}
			int col = (int)floor((x - LEFT)/ faceLength);
			int row = (int)floor(absD(((z - FAR)/faceWidth)));			
			double fractX = (x - LEFT)/ faceLength - col;
			double fractZ = (z - FAR)/faceWidth - row;
			int offset;
			if(fractX + fractZ > 1)
			{
				offset = 2*(this->nHorizontalPoints -1) * row + col * 2 + 1;
			}
			else
			{
				offset = 2*(this->nHorizontalPoints -1) * row + col * 2;
			}	
			return &allFaces.at(offset);
	}
	void getHeightAndNormal(double x,double z,double * y,Arrow * normal)
	{
			if(x < LEFT || x > RIGHT || z < FAR || z > NEAR)
			{
				normal->set(0,1,0);
				*y =0;
				return;
			}

			int col = (int)floor((x - LEFT)/ faceLength);
			int row = (int)floor(absD(((z - FAR)/faceWidth)));
			
			double fractX = (x - LEFT)/ faceLength - col;
			double fractZ = (z - FAR)/faceWidth - row;
			Vertex * v1;
			Vertex * v2;
			Vertex * v3;
			if(fractX + fractZ > 1)
			{
				v1 = &allPoints[col + 1][row];
				v2 = &allPoints[col + 1][row + 1];
				v3 = &allPoints[col][row + 1];
			}
			else
			{
				v1 = &allPoints[col][row];
				v2 = &allPoints[col + 1][row];
				v3 = &allPoints[col][row + 1];
			}
			Arrow a1 = v1->sub(v2);
			Arrow a2 = v3->sub(v2);
			* normal = a1.cross(a2);
			normal->setLength(1);
			double r= normal->dot(v1);
			* y = (r - x*normal->getX() - z*normal->getZ())/normal->getY();			
	}

	void draw()		//draw things
	{
		if(dlSaved)
		{
			glCallList(terrainDL);
		
		}
		else
		{
			terrainDL = glGenLists(1);
			glNewList(terrainDL,GL_COMPILE);
			glColor3f(0.7,0.7,0.7);
			if(bDoFlatShading)								
				glShadeModel(GL_FLAT);
			else
				glShadeModel(GL_SMOOTH);
			    
			if(bPerFaceNormals)
				drawPerFace();
			else
				drawPerVertex();
			glEndList();
			dlSaved =true;
		}

		if(bDrawNormals)
			drawNormals();
	}
	void drawPerFace()	//draws when normals are calculates per face
	{
		glBegin(GL_TRIANGLES);
		int size = (int)allFaces.size();
		Face * p = &allFaces[0];
		glColor3f(0.8,0.8,0.8);
		for(int i = 0;i < size;i++)
		{
			glNormal3dv(p->normal.getCoords());
			glVertex3dv(p->getVertex(0)->getCoords());
			glVertex3dv(p->getVertex(1)->getCoords());
			glVertex3dv(p->getVertex(2)->getCoords());
			p++;
		}
		glEnd();
	}
	void drawPerVertex()	//draws when normals are calculated per vertex
	{
		glEnable(GL_TEXTURE_2D);
		int size = (int)allFaces.size();
		Face * p = &allFaces[0];
		Corner * c1;
		Corner * c2;
		Corner * c3;
		glBindTexture(GL_TEXTURE_2D, texture[0].TexID);
		glBegin(GL_TRIANGLES);
		glColor3f(0.8,0.8,0.8);
		for(int i = 0;i < size / 2;i++)
		{
			//first face
			c1 = (Corner *)(p->getVertex(0));
			c2 = (Corner *)(p->getVertex(1));
			c3 = (Corner *)(p->getVertex(2));

			glNormal3dv(c1->normal.getCoords());
			glColor3dv(c1->cornerColor.getCoords());
			glTexCoord2f(1.0f, 0.0f);
			glVertex3dv(c1->getCoords());

			glNormal3dv(c2->normal.getCoords());
			glColor3dv(c2->cornerColor.getCoords());
			glTexCoord2f(0.0f, 0.0f);
			glVertex3dv(c2->getCoords());

			glNormal3dv(c3->normal.getCoords());
			glColor3dv(c3->cornerColor.getCoords());
			glTexCoord2f(0.0f, 1.0f);
			glVertex3dv(c3->getCoords());
			p++;
			//second face
			c1 = (Corner *)(p->getVertex(0));
			c2 = (Corner *)(p->getVertex(1));
			c3 = (Corner *)(p->getVertex(2));

			glNormal3dv(c1->normal.getCoords());
			glColor3dv(c1->cornerColor.getCoords());
			glTexCoord2f(1.0f, 0.0f);
			glVertex3dv(c1->getCoords());

			glNormal3dv(c2->normal.getCoords());
			glColor3dv(c2->cornerColor.getCoords());
			glTexCoord2f(1.0f, 1.0f);
			glVertex3dv(c2->getCoords());

			glNormal3dv(c3->normal.getCoords());
			glColor3dv(c3->cornerColor.getCoords());
			glTexCoord2f(0.0f, 1.0f);
			glVertex3dv(c3->getCoords());
			p++;
		}
		glEnd();
		glDisable(GL_TEXTURE_2D);
		
	}
	void drawNormals()	
	{
		glLineWidth(2);
		if(bPerFaceNormals) //draws normals for faces
		{
			glColor3f(0.5,1,0.7);
			glBegin(GL_LINES);
			int numFaces = (int)allFaces.size();
			Face * face = &allFaces[0];
			Arrow * normal;
			double * startCoords;
			double endCoords[3];
			double * direction;
			for(int i =0; i< numFaces;i++)
			{
				normal = &face->normal;
				startCoords = face->getVertex(0)->getCoords();
				direction = normal->getCoords();
				endCoords[0] = startCoords[0] + direction[0] * 100;
				endCoords[1] = startCoords[1] + direction[1] * 100;
				endCoords[2] = startCoords[2] + direction[2] * 100;
				glVertex3dv(startCoords);
				glVertex3dv(endCoords);;
				face++;
			}
			glEnd();			
		}
		else	//draws normals for vertices
		{
			glColor3f(0.5,0.7,1);
			glBegin(GL_LINES);
			int numFaces = (int)allFaces.size();
			Face * face = &allFaces[0];
			Arrow * normal;
			double * startCoords;
			double endCoords[3];
			double * direction;
			for(int i =0; i < nHorizontalPoints;i++)
			{
				for(int j =0;j < nVerticalPoints;j++)
				{
					normal = allPoints[i][j].getNormal();
					startCoords = allPoints[i][j].getCoords();
					direction = normal->getCoords();
					endCoords[0] = startCoords[0] + direction[0] * 100;
					endCoords[1] = startCoords[1] + direction[1] * 100;
					endCoords[2] = startCoords[2] + direction[2] * 100;
					glVertex3dv(startCoords);
					glVertex3dv(endCoords);
				}
			}
			glEnd();		
		}
		glLineWidth(1);
		
	}
	void addHill()
	{
			double zDistance = NEAR - FAR;
			double xDistance = RIGHT - LEFT;
			double maxY  = maxBubbleSize;	
			double minY = minBubbleSize;
			double radius = myRandom( minY, maxY);			
			// pick a centerpoint for the hill
			double x, z;

			//big square
			//x = myRandom( LEFT -radius, RIGHT + radius );
			///z = myRandom( FAR -radius, NEAR + radius );
			
			//smoothed square
			x = myRandom( LEFT  + radius * 1.2, RIGHT - radius * 1.2);
			z = myRandom( FAR + radius * 1.2, NEAR - radius * 1.2);

			//island
			/*double theta = myRandom( 0, 6.28 );
			double d = myRandom( radius/3, zDistance /2 - radius/3);		

			x = xDistance /2 + cos( theta) * d;
			z =zDistance /-2 + sin( theta) * d;
			*/
			double radiusSq = radius * radius;
			double distanceSq;
			double height;
			
			// find the range of cells affected by this hill
			int xMin = (int)floor(((x - radius) - LEFT) / this->faceLength);
			int xMax = (int)floor(((x + radius) - LEFT)/this->faceLength);
			// don't affect cell outside of bounds
			if( xMin < 0 ) 
				xMin = 0;
			if( xMax >= this->nHorizontalPoints) 
				xMax = nHorizontalPoints - 1;
			
			int zMin = (int)floor(((z - radius) - FAR) / this->faceWidth);
			int zMax = (int)floor(((z + radius) - FAR)/this->faceWidth);
			if( zMin < 0 ) 
				zMin = 0;
			if( zMax >= this->nVerticalPoints) 
				zMax = nVerticalPoints - 1;
			double actualX,actualZ,prevY;
			for(int h = xMin; h <= xMax; h++ )
			{
				for( int v = zMin; v <= zMax; v++ )
				{
					actualX = allPoints[h][v].getX();
					actualZ = allPoints[h][v].getZ();
					prevY = allPoints[h][v].getY();
					distanceSq = ( x - actualX ) * ( x - actualX ) + ( z - actualZ ) * (  z- actualZ );
					height = radiusSq - distanceSq;
					
					if( height > 0 )
					{
						this->allPoints[h][v].setY(prevY + height);
					}	
				}
			}	
	}
	void normalize()
	{		
			double min= 1000;
			double max = -1000;
			
			// find the min and max
			for( int x = 0; x < nHorizontalPoints; ++x )
			{
				for( int z = 0; z < nVerticalPoints; ++z )
				{
					double y = allPoints[x][z].getY();
					if( y < min) min= y;
					if( y > max )max = y;
				}
			}
			double newRange = maxHeight - minHeight;
			// avoiding divide by zero (unlikely with floats, but just in case)
			double range = max - min;
			if( min != max)
			{
				// divide every height by the maximum to normalize to ( 0.0, 1.0 )
				for( int x = 0; x < nHorizontalPoints; ++x )
				{
					for( int z = 0; z < nVerticalPoints; ++z )
					{
						double y = allPoints[x][z].getY();
						y = (y- min)/range; //from zero to 1
						y = pow(y,smoothingPower);
						y = y * newRange;
						allPoints[x][z].setY(y);
					}
				}
			}	
	}
	void regenerate()	//(re)generates terrain
	{
		dlSaved = false;
		int xSteps,zSteps;
		//generate points
		allFaces.resize((nHorizontalPoints -1)* (nVerticalPoints - 1) * 2);
		double xStep = (RIGHT - LEFT) / (nHorizontalPoints -1);
		double zStep = (NEAR - FAR )/ (nVerticalPoints -1);		
		faceLength = xStep;
		faceWidth=  zStep;
		double xVal;
		double zVal;
		double yVal;
		for(xSteps = 0;xSteps < nHorizontalPoints; xSteps++)
		{
			for(zSteps = 0;zSteps < nVerticalPoints;zSteps++)
			{
    				//creates a point
				xVal = xSteps * xStep + LEFT;
				zVal = zSteps * zStep + FAR;
				yVal = 0;
				allPoints[xSteps][zSteps].set(xVal,yVal,zVal); 
				//resets the attached faces
				allPoints[xSteps][zSteps].clearFaces();
			}
			
		}
		for(int i =0;i < numSpheres;i++)
		{
			addHill();
		}
		this->maxBubbleSize /=2;
		
		for(int i =0;i < numSpheres;i++)
		{
			addHill();
		}


		this->normalize();

		//calculate per face normals
		Corner * c1;
		Corner * c2;
		Corner * c3;
		Corner * c4;
		Face * face = &allFaces[0];
		for(xSteps = 0;xSteps < (nHorizontalPoints - 1);xSteps++)
		{
			for(zSteps = 0;zSteps < (nVerticalPoints - 1);zSteps++)
			{
				
				c1 = &allPoints[xSteps][zSteps];
				c2 = &allPoints[xSteps + 1][zSteps];
				c3 = &allPoints[xSteps + 1 ][zSteps +1];
				c4 = &allPoints[xSteps][zSteps + 1];
				
				//create faces in a given square
				* face = Face();
				face->addVertex(c1);
				face->addVertex(c2);
				face->addVertex(c4);					
				face->calculateNormal();
				c1->addFace(face);
				c2->addFace(face);
				c4->addFace(face);
				face++;
				
				* face = Face();
				face->addVertex(c2);
				face->addVertex(c3);
				face->addVertex(c4);								
				face->calculateNormal();				
				c3->addFace(face );
				c2->addFace(face);
				c4->addFace(face);
				face ++;
			
				//add pointers of created faces to corresponding vertices

			}
		}
		//calculate per vertex normals and per vertex colors
		for(xSteps = 0;xSteps < nHorizontalPoints;xSteps++)
		{
			for(zSteps = 0;zSteps < nVerticalPoints;zSteps++)
			{
				allPoints[xSteps][zSteps].calculateNormal();// calculate normal based on adjacent faces
				allPoints[xSteps][zSteps].calculateColor(); //calculate color based on adjacent faces
			}
		}
	}
};
#endif


